from contextlib import closing
from urllib.parse import urlsplit
from requests import exceptions as rex
import re
import attr

@attr.s
class Hit(object):
    def __bool__(self):
        return self.confidence > 0.0

    @property
    def details(self):
        strings = []
        if self.name:
            strings.append(self.name)
        if self.version:
            strings.append(self.version)
        if self.components:
            strings.append('+'.join(self.components))
        if self.confidence < 1.0:
            strings.append('%d%%' % (self.confidence*100))
        return ', '.join(strings)

    confidence = attr.ib(default=1.0, validator=attr.validators.instance_of(float))
    name = attr.ib(default=None)
    version = attr.ib(default=None)
    components = attr.ib(default=None)

def _meaningless(x, *vals):
    if x not in vals:
        return x

#####
# Sniffers based on protocol details
#####

def global_protect(sess, server):
    '''PAN GlobalProtect'''
    # with closing(sess.get('https://{}/ssl-tunnel-connect.sslvpn'.format(server), stream=True)) as r:
    #    if r.status_code==502:
    #        components.append('gateway')

    components = []
    version = hit = None

    for component, path in (('portal','global-protect'), ('gateway','ssl-vpn')):
        r = sess.post('https://{}/{}/prelogin.esp?tmp=tmp&clientVer=4100&clientos=Windows'.format(server, path), headers={'user-agent':'PAN GlobalProtect'})
        if r.headers.get('content-type','').startswith('application/xml') and b'<prelogin-response>' in r.content:
            hit = True

            if b'<status>Success</status>' in r.content:
                components.append(component)
            elif b'<status>Error</status>' in r.content and b'<msg>Valid client certificate is required</msg>' in r.content:
                components.append(component)
                components.append(component+' wants ccert')

            m = re.search(rb'<saml-auth-method>([^<]+)</saml-auth-method>', r.content)
            if m:
                saml = '%s wants SAML %s' % (component, m.group(1).decode())
                components.append(saml)

            m = re.search(rb'<panos-version>([^<]+)</panos-version>', r.content)
            if m:
                version = m.group(1).decode()

    if hit:
        return Hit(components=components, version=_meaningless(version, '1'))

def check_point(sess, server):
    '''Check Point'''
    confidence = version = None

    # Try an empty client request in Check Point's parenthesis-heavy format
    r = sess.post('https://{}/clients/abc'.format(server), headers={'user-agent':'TRAC/986000125'}, data=b'(CCCclientRequest)')
    if r.content.startswith(b'(CCCserverResponse'):
        confidence = 1.0

    r = sess.get('https://{}/'.format(server), headers={'user-agent':'TRAC/986000125'})
    m = re.search(rb'(\d+-)?(\d+).+Check Point Software Technologies', r.content)
    if m:
        version = m.group(2).decode()
        confidence = confidence or 0.2

    return confidence and Hit(version=version, confidence=confidence)

def sstp(sess, server):
    '''SSTP'''
    # Yes, this is for real...
    # See section 3.2.4.1 of v17.0 doc at https://msdn.microsoft.com/en-us/library/cc247338.aspx

    with closing(sess.request('SSTP_DUPLEX_POST', 'https://{}/sra_%7BBA195980-CD49-458b-9E23-C84EE0ADCD75%7D/'.format(server), stream=True)) as r:
        if r.status_code==200 and r.headers.get('content-length')=='18446744073709551615':
            version = _meaningless( r.headers.get('server'), "Microsoft-HTTPAPI/2.0" )
            return Hit(version=version)

def anyconnect(sess, server):
    '''AnyConnect/OpenConnect'''

    components = []

    # Use XML-post auth to check for client cert requirement
    try:
        r = sess.post('https://{}/'.format(server),
                      headers={'X-Aggregate-Auth':'1', 'X-Transcend-Version':'1'}, data=
                      '<?xml version="1.0" encoding="UTF-8"?>\n'
                      '<config-auth client="vpn" type="init">'
                      '<version who="vpn"/><device-id/>'
                      '<group-access>{}</group-access></config-auth>'.format(server))
        if b'<client-cert-request' in r.content:
            components.append('wants ccert')
    except rex.ChunkedEncodingError:
        pass # some servers barf on this

    with closing(sess.request('CONNECT', 'https://{}/CSCOSSLC/tunnel'.format(server), headers={'Cookie': 'webvpn='}, stream=True)) as r:
        # Cisco returns X-Reason in response to bad CONNECT-tunnel request (GET works too)...
        if 'X-Reason' in r.headers:
            return Hit(name="Cisco", version=r.headers.get('server'), components=components)
        elif r.reason=='Cookie is not acceptable':
            return Hit(name="ocserv", version='0.11.7+', components=components)
        # ... whereas ocserv 7e06e1ac..3feec670 inadvertently sends X-Reason header in the *body*
        elif r.raw.read(9)==b'X-Reason:':
            return Hit(name="ocserv", version='0.8.0-0.11.6', components=components)

def juniper_pulse(sess, server):
    '''Juniper/Pulse'''

    with closing(sess.get('https://{}/'.format(server), headers={'Content-Type':'EAP', 'Upgrade':'IF-T/TLS 1.0', 'Content-Length': '0'}, stream=True)) as r:
        if r.status_code == 101:
            return Hit(name='Pulse Secure', version=r.headers.get('NCP-Version'))

    confidence = None
    r = sess.get('https://{}/dana-na'.format(server), headers={'user-agent':'ncsrv', 'NCP-Version': '3'})
    if any(c.name.startswith('DS') for c in sess.cookies):
        confidence = 1.0
    elif urlsplit(r.url).path.startswith('/dana-na/auth/'):
        confidence = 0.8

    return confidence and Hit(name='Juniper NC', confidence=confidence, version=r.headers.get('NCP-Version'))

#####
# Sniffers based on behavior of web front-end
#####

def openvpn(sess, server):
    '''OpenVPN'''
    r = sess.get('https://{}/'.format(server))
    if any(c.name.startswith('openvpn_sess_') for c in sess.cookies):
        return Hit(version=r.headers.get('server'))

def barracuda(sess, server):
    '''Barracuda'''

    r = sess.get('https://{}/'.format(server))

    m = re.search(rb'(\d+-)?(\d+)\s+Barracuda Networks', r.content)
    version = m and m.group(2).decode()

    confidence = None
    if 'SSLX_SSESHID' in sess.cookies:
        confidence = 1.0
    elif urlsplit(r.url).path.startswith('/default/showLogon.do'):
        confidence = 0.9 if version else 0.8
    elif version:
        confidence = 0.2

    return confidence and Hit(version=version, confidence=confidence)

def fortinet(sess, server):
    '''Fortinet'''

    # server sets *empty* SVPNCOOKIE/SVPNNETWORKCOOKIE
    r = sess.get('https://{}/remote/login'.format(server))
    if r.headers.get('set-cookie','').startswith('SVPNCOOKIE'):
        server = r.headers.get('server')
        confidence = 1.0 if server=='xxxxxxxx-xxxxx' else 0.9
        return Hit(confidence=confidence, version=_meaningless(server,'xxxxxxxx-xxxxx'))

def array_networks(sess, server):
    '''Array Networks'''

    r = sess.get('https://{}/'.format(server))

    confidence = 0
    if re.match(r'/prx/\d\d\d/', urlsplit(r.url).path):
        confidence += 0.1
        if b'array networks' in r.content.lower() or b'arraynetworks' in r.content.lower():
            confidence += 0.1
        if b'_AN_global_var_init' in r.content:
            confidence += 0.2

    return confidence and Hit(confidence=confidence)

sniffers = [
    anyconnect,
    juniper_pulse,
    global_protect,
    barracuda,
    check_point,
    sstp,
    openvpn,
    fortinet,
    array_networks,
]



#!/usr/bin/env python3

import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from requests import exceptions as rex
from .sniffers import sniffers, Hit
from .requests import SnifferSession
from .version import __version__
import socket
import logging
import http.client
from ipaddress import ip_address

import argparse
import os
import re
import csv
from sys import stderr, stdout

def main():
    p = argparse.ArgumentParser()
    p.add_argument('-k','--keep-going', action='store_true', help='Keep going after first hit')
    p.add_argument('-t','--timeout', metavar='SEC', type=lambda x: int(x) or None, default=10, help='Timeout in seconds (default %(default)s, 0 for none)')
    x = p.add_mutually_exclusive_group()
    x.add_argument('-v','--verbose', default=0, action='count')
    x.add_argument('-c','--csv', action='store_true', help='Output report in CSV format')
    p.add_argument('server', nargs='+', help='suspected SSL-VPN server')
    p.add_argument('-L','--logging', action='store_true', help='Detailed logging for requests and httplib')
    p.add_argument('-V','--version', action='version', version='%(prog)s ' + __version__)
    args = p.parse_args()

    if args.logging:
        http.client.HTTPConnection.debuglevel = 99
        logging.basicConfig(level=logging.DEBUG)

    if args.csv:
        wr = csv.writer(stdout)
        wr.writerow(('Server','Errors','Sniffer','Confidence','Name','Version','Components'))

    s = SnifferSession()
    s.timeout = args.timeout

    for server in args.server:
        if args.verbose:
            print("\nSniffing {} ...".format(server))
        elif not args.csv:
            print("{}: ".format(server), end='')

        if server.startswith('[') and ']' in server:
            # IPv6 address?
            domain = server[1:].split(']', 1)[0]
        else:
            domain = server.split(':', 1)[0]
        try:
            ip_address(domain)
        except ValueError:
            try:
                socket.gethostbyname(domain)
            except socket.error:
                if args.csv:
                    wr.writerow((server, "DNS lookup failed"))
                else:
                    print("DNS lookup failed")
                continue

        hits = []
        ssle = timeout = 0
        for sniffer in sniffers:
            desc = sniffer.__doc__ or sniffer.__name__
            if args.verbose:
                print("  Is it {}? ".format(desc), end='')

            s.cookies.clear()
            hit = ex = None
            try:
                hit = sniffer(s, server)
            except rex.Timeout as e:
                ex = 'timeout'
                timeout += 1
            except rex.SSLError as e:
                ex = 'SSL error'
                ssle += 1
            except rex.ConnectionError as e:
                ex = 'connection error'
            except Exception as e:
                ex = e.__class__.__name__
            else:
                ex = 'no match'

            if hit:
                if args.csv:
                    wr.writerow((server, None, sniffer.__name__, hit.confidence, hit.name, hit.version, hit.components and '+'.join(hit.components)))
                elif hit.details:
                    desc += ' ({})'.format(hit.details)
                hits.append(desc)
            if args.verbose:
                print((hit.details or 'hit') if hit else ex)

            if hit and not args.keep_going:
                break

        if args.verbose:
            print("  => ", end='')

        if ssle:
            errs = 'SSL errors'
        elif timeout:
            errs = 'timeout'
        elif not hits:
            errs = 'no match'

        if args.csv and not hits:
            wr.writerow((server, errs))
        elif not args.csv:
            print(', '.join(hits) or errs)

########################################

if __name__=='__main__':
    main()



import requests

class TimeoutSession(requests.Session):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.timeout = None
    def request(self, *a, **kw):
        kw.setdefault('timeout', self.timeout)
        return super().request(*a, **kw)

class SnifferSession(TimeoutSession):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        del self.headers['user-agent']
        self.verify = False



#!/usr/bin/env python3

import sys, os
try:
  from setuptools import setup
except ImportError:
  from distutils.core import setup

if not sys.version_info[0] == 3:
    sys.exit("Python 2.x is not supported; Python 3.x is required.")

########################################

version_py = os.path.join('what_vpn', 'version.py')

d = {}
with open(version_py, 'r') as fh:
    exec(fh.read(), d)
    version_pep = d['__version__']

########################################

setup(name="what-vpn",
      version=version_pep,
      description="Identify servers running various SSL VPNs",
      long_description=open("description.rst").read(),
      author="Daniel Lenski",
      author_email="dlenski@gmail.com",
      license='GPL v3 or later',
      install_requires=[ 'requests>=2.0.0' ],
      url="https://github.com/dlenski/what-vpn",
      packages = ['what_vpn'],
      entry_points={ 'console_scripts': [ 'what-vpn=what_vpn.__main__:main' ] },
      test_suite='nose.collector',
      )
